* QL WORLD DIY TOOLKIT - EDLINE$ function, by Simon N Goodwin
* rewritten by Laurence W Reeves.
* V0.03, pjw, 2022 Mar 01, Many changes:
*       SMSQ/E mnemonics, SMSQ/E includes, Qmac-compatible
*       Split init into separate file for Pick'n'mix toolkit
*       String returned in parameter; terminating key code return
*       (ie probably not Turbo-compatible)
* V0.04, pjw, 2022 Apr 02, pixel positioning. Stack bug on ESC fixed

        section edline

        include dev8_keys_qdos_io
        include dev8_keys_qlv
        include dev8_keys_sbasic
        include dev8_keys_err

        xdef edline

edline
        moveq   #err.ipar,d0      BAD PARAMETER error
        add.w   #3*8,a3
        cmp.l   a3,a5
        bne     rts0            Ensure exactly three parameters

* Get the default string

        subq.l  #8,a3           Retrieve just the last parameter
        move.w (a6,a3.l),d1     get type
        andi.w #$0f0f,d1
        subq.w #1,d1            unset string is ok
        beq.s ok

        subi.w #$200,d1         set string ok too
        bne.s  rts0

ok
        movea.l a3,a4           a4: keep a3 safe for later

        move.w  sb.gtstr,a2     get a string
        jsr     (a2)
        bne.s   rts0

* Read the first two parameters: channel number and buffer size

        move.l  a3,a5
        sub.w   #2*8,a3         get the other two parameters
        move.w  sb.gtint,a2     get two ints
        jsr     (a2)
        bne.s   rts0
        movem.w 0(a1,a6.l),d5-d7 Get channel number, buffer and string lengths

* Convert channel number to channel ID

        moveq   #err.ichn,d0    CHANNEL NOT OPEN error
        mulu    #ch.len,d5      Find offset in table
        add.l   sb_chanb(a6),d5 Add base offset
        cmp.l   sb_chanp(a6),d5 Check not beyond end
        bge.s   rts0
        move.l  ch_chid(a6,d5.l),d5     Get channel ID
        bmi.s   rts0            Channel must be open

* Validate initial string length

        moveq   #err.bffl,d0    BUFFER FULL error
        move.l  d6,d1
        sub.w   d7,d1           is the string already too big for the buffer?
        blt.s   rts0            if so (or buffer length is negative), get out

        addq.l  #2,d1           allow extra for coordinates if string is null
        move.w  qa.resri,a2
        jsr     (a2)            ensure enough space for longest text

        move.l  sb_arthp(a6),a2
        addq.l  #6,a2           position to start of default string

* Find out where we are on the screen

        move.l  d5,a0           Channel ID
        moveq   #-1,d3          Timeout

        lea     -8(a2),a1       place to read initial coords to
        moveq   #iow.pixq,d0
        bsr.s   rel_QDOS
        move.l  -4(a6,a2.l),d5  pick up coords and save them

* Put the string in the buffer

        addq.l  #1,d6           add one to buffer length for terminator
        lea     0(a2,d7.l),a3
        sub.l   d6,a3           good enough for buffer location
        move.l  a3,a1
        move.w  d7,d1
        bra.s   buff_ent

rel_QDOS
        trap    #4
call_QDOS
        trap    #3              General-purpose ROM caller
        tst.l   d0
        beq.s   rts0
        addq.l  #4,a7           Return to previous caller
rts0
        rts

buff_lp
        move.b  0(a6,a2.l),0(a6,a1.l)
        addq.l  #1,a2
        addq.l  #1,a1
buff_ent
        dbra    d1,buff_lp
        moveq   #1,d1
        and.w   d7,d1
        add.w   d1,a2           set top of stack

* Go to the initial cursor position and print the text

go_there
        move.l  d5,d1
        swap    d1              Extract cursor X
        move.w  d5,d2           Extract cursor Y
        moveq   #iow.spix,d0
        bsr.s   call_QDOS
        move.l  a3,a1
        move.w  d7,d2
        moveq   #iob.smul,d0
        bsr.s   rel_QDOS
        moveq   #' ',d1
        moveq   #iob.sbyt,d0    blank end of line
        bsr.s   call_QDOS
        moveq   #iow.pcol,d0    move cursor to blank
        bsr.s   call_QDOS

* Call up the editor

        move.w  d7,d1           Put cursor at end of text
        swap    d1
        move.w  d7,d1           Indicate length of text
        lea     0(a3,d1.w),a1   Point to last character
        move.w  d6,d2           Maximum length
        moveq   #iob.elin,d0
        trap    #4
        trap    #3
        subq.w  #1,d1           Don't count last character
        move.w  d1,d7           Slide up to the end-stop
        move.l  d0,d2           Save and test the error code
        beq.s   edited          No error: check terminator
        moveq   #iow.dcur,d0    for AH & JM ROMs & err.bffl
        bsr.s   call_QDOS       which leave the cursor on
        move.l  d2,d0           restore the error code
        addq.l  #-err.bffl,d2   Was it buffer overflow?
        beq.s   go_there        Yes, so try again
        rts                     Give up if anything else!

edited
;       sub.b   #10,-1(a6,a1.l) (tidy to make pad byte come out zero)
;       bne.s   go_there        Restart unless last character was ENTER

        move.b  -1(a6,a1.l),d5  d5 = terminating char
        cmpi.b #27,d5
        beq.s esc

        btst    d0,d1           see if string is an odd length
        beq.s   fin_ent         no padding byte
fin_loop
        move.b  -1(a6,a1.l),-1(a6,a2.l)
        subq.l  #1,a2
fin_ent
        subq.l  #1,a1
        dbra    d7,fin_loop

        move.l  a2,a1
        subq.l  #2,a1           Make room to stack length 
        move.l  a1,sb_arthp(a6) Reset maths stack pointer
        move.w  d1,0(a6,a1.l)   set string length

        movea.l a4,a3           a3-> string parameter
        move.w sb.putp,a2       stuff string into parameter
        jsr (a2)                "empties" ri stack

        bra.s ende
esc
        move.l a2,sb_arthp(a6)  reset ri stack

ende
        subq.l #2,sb_arthp(a6)  make room for int return
        movea.l sb_arthp(a6),a1
        move.w d5,(a6,a1.l)
        moveq #ar.int,d4        Return an int

        rts

        end
